<?php

/**
 * Class for import resources from Open Journal System
 */
class ElisImportOjs {
  public $journalUrl = false;
  public $journalNode = false; // (node) of root journal
  public $sync = false; // true if resync exist node, false - add only new node. Sync metadata only, not files.
  public $from = false; // Y-m-d from request. Example 1980-01-01. false - not set
  public $until = false; // Y-m-d until request. Example 2025-01-01. false - not set
  
  public function __construct($journalNid) {
    $node = node_load($journalNid);
    
    $journalUrl = ElisHelper::getFirstFieldValue('node', $node, 'field_import_source');
    print $journalUrl . "\n";
    if ($node != false && $journalUrl != false && valid_url($journalUrl, true) && self::isJournalNode($node)) {
      $this->journalNode = $node;
      $this->journalUrl = $journalUrl;
    }
    
  }
  
  public function getEndPoint() {
    return $this->journalUrl . '/oai';
  }
  
  private function getIdentsFilePath() {
    return sys_get_temp_dir() . '/elis_ojs_idents_' . $this->journalNode->nid . '.tsv';
  }
  
  private function getProblemsFilePath() {
    return sys_get_temp_dir() . '/elis_ojs_problems_' . $this->journalNode->nid . '.tsv';
  }
  
  /**
   * Import resources from journalUrl to nid folder with structure
   * @param string $journalUrl. Example: https://press.psu.ru/index.php/history
   * @param int $nid - node id for import journal
   */
  public function importJournal() {   
    $this->getAllIdentsFromOaiToFile();
    $this->importJournalFromIdentsFile();
    
  }
  
  public function getFromUntilRequestString() {
    $str = '';
    if (isset($this->from) && preg_match("/^\d\d\d\d-\d\d-\d\d/", $this->from) && isset($this->until) && preg_match("/^\d\d\d\d-\d\d-\d\d/", $this->until)) {
      $str = '&from=' . $this->from . '&until=' . $this->until;
    }
    return $str;
  }
  
  /**
   * Get all records from OAI-PMH and save it to file in /tmp/elis_oai_idents_$nid.tsv and save problems url into /tmp/elis_oai_problems_$nid.tsv
   */
  public function getAllIdentsFromOaiToFile() {
    $resumptionToken = variable_get('ojs_oaipmh_import_resumptiontoken', null); // for continue
    $resumptionToken = null;  // for reset

    $file_ident_path = $this->getIdentsFilePath(); // file with identifiers
    $file_problem_path = $this->getProblemsFilePath(); // file with problem idents

    if (file_exists($file_ident_path)) {
      unlink($file_ident_path); // clean old idents
    }
    //unlink($file_problem_path);
    
    print "load with resumption token: '{$resumptionToken}'\n";
    // first step is initialized metadataPrefix or if resumptionToken is not null, then get next iteration
    do {
      if ($resumptionToken == null) { // initial get with metadataPrefix and without resumptionToken
        $url = $this->getEndPoint() . '?verb=ListRecords&metadataPrefix=oai_dc' . $this->getFromUntilRequestString();
        print "$url\n";
        if (!file_exists($file_ident_path))
          file_put_contents($file_ident_path, "ident\turl\n", FILE_APPEND);
        if (!file_exists($file_problem_path))
          file_put_contents($file_problem_path, "ident\turl\n", FILE_APPEND);

      } else {
        $url = $this->getEndPoint() . '?verb=ListRecords&resumptionToken=' . urlencode($resumptionToken) . $this->getFromUntilRequestString();
      }
      
      $records_with_url = array();
      $warning_records = array();
      
      $xml_string = file_get_contents($url);
      $xml = simplexml_load_string($xml_string);
      
      if (isset($xml->error)) {
        //print $xml->error['code'] . ' ' . $xml->error->__toString();
        watchdog('elis_ojs', 'get url ":url" records error: ":error"',array(":url"=>$url,":error"=>$xml->error->__toString()));
        break;
      }
      
      $resumptionToken = self::getResumptionTokenFromXml($xml);
      
      $records = self::getRecordsFromXml($xml);
      
      foreach ($records as $i=>$rec) {
        $header = $rec->header;
        $metadata = $rec->metadata;
        //$rec->registerXPathNamespace('oai_dc','http://www.openarchives.org/OAI/2.0/oai_dc/');
        $rec->registerXPathNamespace('dc','http://purl.org/dc/elements/1.1/');
        
        $error = true;
        try {
          if (null != $rec->xpath(".//dc:format[text() = 'application/pdf']")) { // is pdf
        
            $fields = array();
        
            $fields['title'] = $rec->xpath('.//dc:title');
            $fields['field_authors'] = $rec->xpath('.//dc:creator');
            $fields['field_tags'] = $rec->xpath('.//dc:subject');
            $fields['body'] = $rec->xpath('.//dc:description');
            $fields['field_publisher'] = $rec->xpath('.//dc:publisher');
            $fields['field_publish_date'] = $rec->xpath('.//dc:date'); // 2020-03-24
            $record_type = $rec->xpath('.//dc:date'); // can be ['info:eu-repo/semantics/article','info:eu-repo/semantics/publishedVersion']
            $record_format = $rec->xpath('.//dc:format'); // can be ['application/pdf']
            $fields['field_identifier'] = $rec->xpath('.//dc:identifier'); // can be double ['https://ntb.gpntb.ru/jour/article/view/550','10.33186/1027-3689-2020-3-23-36']
            $relations = $rec->xpath('.//dc:relation'); // many relations. First is link to html page with viewer 'https://ntb.gpntb.ru/jour/article/view/550/514' and second is string of literacy
            //$url = $rec->xpath(".//dc:identifier['starts-with(text(),'" . $this->journalUrl . "')]");
            if ($fields['field_identifier'] != null && isset($fields['field_identifier'][0])) { // find url of page with article
              $url_article = null;
              foreach ($fields['field_identifier'] as $field) {
                $field_string = $field->__toString();
                if (strpos($field_string, $this->journalUrl) === 0) {
                  $url_article = $field_string;
                }
              
              }
              if ($url_article != null) {
                $records_with_url[] = $rec;
                file_put_contents($file_ident_path, "{$rec->header->identifier}\t{$url_article}\n", FILE_APPEND);
                $error = false;
              }
            }
            
          }
        } catch (Exception $err) {
          $error = true;
        }
        if ($error === true) {
          $warning_records[] = $rec;
          file_put_contents($file_problem_path, "{$rec->header->identifier}\t{$url_article}\n", FILE_APPEND);
        }
      }
      
      print "count records: " . count($records_with_url) . " " . count($warning_records) . "\n";

      variable_set('ojs_oaipmh_import_resumptiontoken', $resumptionToken);

    } while ($resumptionToken != null);
    variable_del('ojs_oaipmh_import_resumptiontoken');
  }
  
  /**
   * import from exists idents file, that generate on prev step
   */
  public function importJournalFromIdentsFile() {
    $identsFile = file_get_contents($this->getIdentsFilePath());
    $input_idents_str = explode("\n",$identsFile);
    $input_idents = array();
    $allowed_langs = (module_exists('locale')) ? array_keys(locale_language_list()) : array();
    foreach ($input_idents_str as $i=>$idents_str) { // parse tsv
      if ($i>0) {
        $value_str = explode("\t", $idents_str);
        $res = new stdClass();
        if (is_string($value_str[0]) && strlen($value_str[0]) && is_string($value_str[1]) && strlen($value_str[1])) {
          $res->oaiIdent = $value_str[0];
          $res->httpUrl = $value_str[1];
          $input_idents[] = $res;
        }
      }
    }
    foreach ($input_idents as $i=>$res) {
      try {
        $node = ElisHelper::findNodeByIdentifier($res->oaiIdent);
        if ($node != false && $this->sync === false) continue; // work only not exists node
        
        $node_exists = ($node != false) ? true : false; // here save node that exists
        
        if (false == $node_exists) {
          $node = new stdClass();
          $node->type = 'pdf';
          node_object_prepare($node);
          ElisImportHelper::attachIdentifierToNode($node, $res->oaiIdent);
        }
        // copy fields to node
        $cloned_fields = array('field_open_access','field_publisher','field_tags','field_classification','field_allow_unprotected_download','field_license', 'field_serial');
        foreach ($cloned_fields as $field_name) {
          if (isset($this->journalNode->$field_name)) {
            $node->$field_name = $this->journalNode->$field_name;
          }
        }
        ElisImportHelper::attachAllowUnprotectedDownloadToNode($node, '1');
        
        $url_oai = $this->getEndPoint() . '?verb=GetRecord&identifier=' . urlencode($res->oaiIdent) . '&metadataPrefix=oai_dc';//  . $this->getFromUntilRequestString();
        $xml_string_oai = file_get_contents($url_oai);
        
        $xml_oai = simplexml_load_string($xml_string_oai);
        $url_marcxml = $this->getEndPoint() . '?verb=GetRecord&identifier=' . urlencode($res->oaiIdent) . '&metadataPrefix=marcxml';// . $this->getFromUntilRequestString();
        $xml_string_marcxml = file_get_contents($url_marcxml);
        $xml_marcxml = simplexml_load_string($xml_string_marcxml);
        $html_string = file_get_contents($res->httpUrl);
        $htmlDom = DOMDocument::loadHTML($html_string);
        
        print "$url_oai\n";
        print "$url_marcxml\n";
        
        if (isset($xml_oai->error)) {
          watchdog('elis_ojs', 'get url ":url" record error: ":error"',array(":url"=>$url_oai,":error"=>$xml_oai->error->__toString()), WATCHDOG_ERROR);
          continue;
        }
        
        $record_oai = $xml_oai->GetRecord->record->metadata;
        $record_oai->registerXPathNamespace('oai_dc','http://www.openarchives.org/OAI/2.0/oai_dc/');
        $record_oai->registerXPathNamespace('dc','http://purl.org/dc/elements/1.1/');
        $record_oai->registerXPathNamespace('xml','http://www.w3.org/XML/1998/namespace');
        $fields = array();
        
        $record_language = $record_oai->xpath('.//dc:language'); // find language from DC
        if ($record_language != null && isset($record_language[0]) && strlen($record_language[0]->__toString())>=2) {
          $record_language = substr($record_language[0]->__toString(), 0, 2);
          $node->language = $record_language;
        } else {
          $record_language = 'und';
        }
        $fields['title'] = $record_oai->xpath('.//dc:title');
        $fields['field_authors'] = $record_oai->xpath('.//dc:creator');
        $fields['field_tags'] = $record_oai->xpath('.//dc:subject');
        $fields['body'] = $record_oai->xpath('.//dc:description');
        $fields['field_publisher'] = $record_oai->xpath('.//dc:publisher');
        $fields['field_publish_date'] = $record_oai->xpath('.//dc:date'); // 2020-03-24
        $record_type = $record_oai->xpath('.//dc:date'); // can be ['info:eu-repo/semantics/article','info:eu-repo/semantics/publishedVersion']
        $record_format = $record_oai->xpath('.//dc:format'); // can be ['application/pdf']
        $fields['field_identifier'] = $record_oai->xpath('.//dc:identifier'); // can be double ['https://ntb.gpntb.ru/jour/article/view/550','10.33186/1027-3689-2020-3-23-36']

        $relations = $record_oai->xpath('.//dc:relation'); // many relations. First is link to html page with viewer 'https://ntb.gpntb.ru/jour/article/view/550/514' and second is string of literacy
        $authors = array();
        
        $language_default = language_default('language');
        
        // find exists node
        if ($fields['field_identifier'] != null) {
          foreach ($fields['field_identifier'] as $ident) {
            if ($node_exists == false) {
              $node_exists = ElisHelper::findNodeByIdentifier($ident->__toString());
            }
          }
        }
        if ($node_exists != false) {
          if ($this->sync === false) continue; // not sync exists node
          $node = $node_exists; // copy metadata from exists node to node 
        }
        
        foreach ($fields as $field_name=>$values) {
          foreach ($values as $value) {
            $value_str = $value->__toString();
            $lang = 'und';
            
            // find lang for current value
            $ns = $value->getNamespaces(true);
            if (isset($ns['xml'])) {
              $attribs = $value->attributes($ns['xml']); 
              foreach ($attribs as $attr_name=>$attr_value) {
                switch ($attr_name) {
                  case 'lang':
                    if (preg_match("/^[a-zA-Z]{2}/", $attr_value)) {
                      $lang = strtolower(substr($attr_value,0,2));
                    }
                    break;
                }
              }
            }
            
            if (count($values) == 1) $lang = 'und'; // only one record save into 'und' lang
            //if ($node->language == $lang) $lang = 'und'; // if site not localized, we must set 'und' lang
            switch ($field_name) {
              case 'title':
               // if ($lang == $language_default) $node->title = $value_str;
                if (!preg_match("/^http/", $value_str)) { // same time title started from http (not known why)
                  $node->title_field[$lang][0]['value'] = $value_str; // if field title exists
                  if ($node->language == $lang) {
                    $node->title = $value_str;
                  }
                }
                break;
              case 'body':
                $node->$field_name[$lang][0]['value'] = $value_str;
                break;
              case 'field_publish_date':
                $node->$field_name['und'][0]['value'] = $value_str;
                break;
              case 'field_publisher':
        //        ElisImportHelper::attachPublisherToNode($node, array($value_str));
                break;
              case 'field_authors':
                if ($node->language == $lang) {
                  ElisImportHelper::attachAuthorsToNode($node, array($value_str));
                }
                break;
              case 'field_tags':
                $field_info_field = field_info_field($field_name);
                $vocabulary_machine_name = $field_info_field['settings']['allowed_values'][0]['vocabulary'];
                $vocabulary = taxonomy_vocabulary_machine_name_load($vocabulary_machine_name);
                ElisImportHelper::attachTermToNodeField($node, $field_name, $value_str, $vocabulary);
                break;
              case 'field_identifier':
                if ($value_str != 'undefined' && strlen($value_str) && $value_str != 'null' && $value_str != '0' && $value_str != 'false') {
                  ElisImportHelper::attachIdentifierToNode($node, $value_str);
                }
                break;
            }
          }
        }
        
        $fileDownloadUrl = false;
        // workd with html
        $htmlXpath = new DOMXPath($htmlDom);
        $metas = $htmlXpath->query("//html/head/meta");
        foreach ($metas as $meta) {
          $metaName = $meta->getAttribute('name');
          $metaContent = $meta->getAttribute('content');
          $metaLang = 'und'; // default language of meta element
          foreach ($meta->attributes as $key=>$attr) { // find metaLang
            switch ($attr->name) {
              case 'xml:lang':
                if (preg_match("/^[a-zA-Z]{2}/", $attr->value)) {
                  $metaLang = strtolower($attr->value);
                  if (strlen($metaLang)>2) {
                    $metaLang = substr($metaLang, 0, 2);
                  }
                }
                break;
            }
          }
          print "$metaLang $metaContent\n";
          //if ($metaLang == $node->language) $metaLang = 'und'; // for node language set lang as 'und' for correct work with not localized drupal
          
          switch ($metaName) {
            case 'DC.Creator.PersonalName':
              //print "author: {$metaContent}\n";
              //ElisImportHelper::attachAuthorsToNode($node, array($metaContent));
              break;
            case 'DC.Identifier.pageNumber':
              $node->field_is_fragments['und'][0]['value'] = $metaContent;
              print "pages: {$metaContent}\n";
              break;
            case 'DC.Title':
              $node->title_field[$metaLang][0]['value'] = $metaContent;
              if (!isset($node->title) && $node->language == $metaLang) {
                $node->title = $metaContent; 
              }
              break;
            case 'DC.Title.Alternative':
              if (!isset($node->title_field[$metaLang]) && $metaLang != 'und') {
                $node->title_field[$metaLang][0]['value'] = $metaContent;
              }
              break;
            case 'DC.Type':
              if ($metaContent == 'Text.Serial.Journal') {
                ElisImportHelper::attachClassificationTermsToNodeIfExists($node, array('Научная статья'));
              }
              break;
            case 'DC.Language':
              $node->language = substr($metaContent,0,2);
              break;
            case 'DC.Description':
              $node->body[$metaLang][0]['value'] = $metaContent;
              break;
            case 'citation_issn':
              break;
            case 'citation_volume':
              if ($metaContent != '0') {
                $node->field_volume['und'][0]['value'] = $metaContent;
              }
              break;
            case 'citation_issue':
              $node->field_issue['und'][0]['value'] = $metaContent;
              break;
            case 'citation_journal_title':
              ElisImportHelper::attachSerialToNode($node, array($metaContent));
              break;
            case 'citation_abstract_html_url':
              $node->field_import_source['und'][0]['value'] = $metaContent;
              break;
            case 'citation_reference':
              $field_info_biblio_references = field_info_field('field_biblio_references');
              if ($field_info_biblio_references != false) {
                $node->field_biblio_references[$metaLang][]['value'] = $metaContent;
              }
              break;
            case 'citation_date':
              $metaContent = preg_replace("/\//","-",$metaContent); // normalize
              $metaContent = preg_replace("/\./","-",$metaContent); // normalize
              if (preg_match("/(\d\d\d\d)-01-01/", $metaContent, $matches)) { // it is fake data. Use only year
                $metaContent = $matches[1];
              }
              $node->field_publish_date['und'][0]['value'] = $metaContent;
              break;
            case 'DC.Identifier.DOI':
              //ElisImportHelper::attachIdentifierToNode($node, $metaContent); // attached before
              break;
            case 'citation_pdf_url':
              // link to pdf file
              $fileDownloadUrl = $metaContent; // need import
              break;
          }
        }
        
        
        
        
        // post processing
        
        // add references if it have in oai_dc, but not exists on page
        
        if (!isset($node->field_biblio_references) && $relations != null) {
          foreach ($relations as $i=>$relation_xmlelement) {
            $relation_value = $relation_xmlelement->__toString();
            if ($i>0) { // first relation is href to self skipped
              $node->field_biblio_references['und'][]['value'] = $relation_value;
            }
          }
        }
        
        if (isset($node->field_serial) && (isset($node->field_volume) || isset($node->field_publish_date)) && isset($node->field_issue)) {
            
          $node_chain = ElisSerial::createCatalogTree($this->journalNode, 
            false, //ElisHelper::getFirstFieldValue('node', $node, 'field_volume'),  // same time volume is incorrect and always is 1
            ElisHelper::getFirstFieldValue('node', $node, 'field_issue'), 
            ElisHelper::getFirstFieldValue('node', $node, 'field_publish_date')
          );
          if (count($node_chain)) {
            $parent_node = $node_chain[count($node_chain)-1];
            ElisImportHelper::attachParentCatalogToNode($node, $parent_node);
            print "parent catalog attached\n";
            //node_save($node);
          }
        }
        
        // translate with entity_translation
        if (module_exists('entity_translation') && ($node->language != 'und')) {
          // find additional lang:
          
          $additional_langs = array();
          foreach ($node->title_field as $lang=>$title_fields) {
            if ($node->language != $lang && $lang != 'und') $additional_langs[] = $lang;
          }
          foreach ($additional_langs as $lang) {
            if (in_array($lang, $allowed_langs)) {
            $node->translations->data[$lang]['entity_type'] = 'node'; 
            $node->translations->data[$lang]['entity_id'] = $node->nid;
            $node->translations->data[$lang]['language'] = $lang; 
            $node->translations->data[$lang]['source'] = $node->language;   
            $node->translations->data[$lang]['uid'] = $node->uid;  
            $node->translations->data[$lang]['status'] = $node->status; 
            $node->translations->data[$lang]['translate'] = 0; 
            $node->translations->data[$lang]['created'] = $node->created; 
            $node->translations->data[$lang]['changed'] = $node->changed;
            }
          }
          if (count($additional_langs)) {
            //node_save($node);
          }
        }
        
        // download file
        if ($fileDownloadUrl != false) {
          $options = array();
          if (is_string($node->title) && strlen($node->title)) {
            $options['title'] = $node->title;
          } else { // node not have title
            if (isset($node->title_field)) {
              if (isset($node->title_field['und']) && isset($node->title_field['und'][0]) && isset($node->title_field['und'][0]['value'])) {
                $node->title = $node->title_field['und'][0]['value'];
              } else if (isset($node->title_field[$node->language]) && isset($node->title_field[$node->language][0]) && isset($node->title_field[$node->language][0]['value'])) {
                $node->title = $node->title_field[$node->language][0]['value'];
              } else {
                foreach ($node->title_field as $lng => $title_field) {
                  if (isset($title_field['value'])) {
                    $node->title = $title_field['value'];
                  }
                }
              }
            }
          }
          $options['node'] = $node;
          if (!isset($node->field_file)) {
            import_with_options_elis_pdf($fileDownloadUrl, $options);
          }
        } elseif (!isset($node->field_file)) {
          watchdog('elis_ojs', 'can not find pdf url for ":url_oai"', array(':url_oai'=>$url_oai), WATCHDOG_ERROR);
        }
        
      } catch (Exception $err) {
        watchdog('elis_ojs', $err->getMessage(), array(), WATCHDOG_ERROR);
      }
    }
  }
  
  /**
   * Check that nid is folder type serial without volume and issue.
   * If journalUrl is setted, alsow check that $node->field_import_source == $journalUrl
   * @param int $nid
   * @param string $journalUrl (example: "https://press.psu.ru/index.php/history") 
   * @return boolean true if is serial and false if not
   */
  private static function isJournalNode($node, $journalUrl = false) {
    if ($node == false) {
      watchdog('elis', 'ELiS Import OJS can not import journal ":journalUrl" because root node == false',array(':journalUrl'=>$journalUrl), WATCHDOG_ERROR);
      return false;
    }
    
    if ($node->type != 'subj') {
      watchdog('elis', 'ELiS Import OJS can not import journal node ":nid" is not subj',array(':nid'=>$node->nid), WATCHDOG_ERROR);
      return false;
    }
    
    $res = ElisSerial::extractSerialDataFromDocumentNode($node);
    
    if ($res->serial != false && $res->volume == false && $res->issue == false) {
      if ($journalUrl == false) {
        return true;
      } else {
        $field_import_source = ElisHelper::getFirstFieldValue('node', $node, 'field_import_source');
        if ($field_import_source == $journalUrl) {
          return true;
        } else {
          watchdog('elis', 'ELiS Import OJS can not import journal: node ":nid" is not have field_import_source',array(':nid'=>$node->nid), WATCHDOG_ERROR);
        }
      }
    }
    return false;
  }
  
  public static function getResumptionTokenFromXml($xml) {
    $resumptionToken = false;
    if (isset($xml->ListRecords) && isset($xml->ListRecords->resumptionToken) && isset($xml->ListRecords->resumptionToken[0])) {
      $resumptionToken = $xml->ListRecords->resumptionToken[0]->__toString();
    }
    return $resumptionToken;
  }
  
  private static function getRecordsFromXml($xml) {
    $records = array();
    if (isset($xml->ListRecords) && isset($xml->ListRecords->record)) {
      foreach ($xml->ListRecords->record as $i=>$rec) {
        $records[] = clone $rec;
      }
    }
    return $records;
  }
}
